# -*- coding: utf-8 -*-

"""
Application micc
"""

import os, sys, shutil
from types import SimpleNamespace
from pathlib import Path

import click

from et_micc.project import Project, micc_version
import et_micc.logger

__template_help = "Ordered list of Cookiecutter templates, or a single Cookiecutter template."


@click.group()
@click.option('-v', '--verbosity', count=True
    , help="The verbosity of the program output."
    , default=1
)
@click.option('-p', '--project-path'
    , help="The path to the project directory. "
           "The default is the current working directory."
    , default='.'
    , type=str
)
@click.option('--clear-log'
    , help="If specified clears the project's ``et_micc.log`` file."
    , default=False, is_flag=True
)
@click.version_option(version=micc_version())
@click.pass_context
def main(ctx, verbosity, project_path, clear_log):
    """Micc command line interface.

    All commands that change the state of the project produce some output that
    is send to the console (taking verbosity into account). It is also sent to
    a logfile ``et_micc.log`` in the project directory. All output is always appended
    to the logfile. If you think the file has gotten too big, or you are no more
    interested in the history of your project, you can specify the ``--clear-log``
    flag to clear the logfile before any command is executed. In this way the
    command you execute is logged to an empty logfile.

    See below for (sub)commands.
    """
    if verbosity > 1:
        print("Current micc command is using Python", sys.version.replace('\n', ' '), end='\n\n')

    if clear_log:
        os.remove(project_path / 'micc.log')

    ctx.obj = SimpleNamespace(
        verbosity=verbosity,
        project_path=Path(project_path).resolve(),
        default_project_path=(project_path=='.'),
        clear_log=clear_log,
        template_parameters={},
    )


@main.command()
@click.option('--publish'
    , help="If specified, verifies that the package name is available on PyPI.\n"
           "If the result is False or inconclusive the project is NOT created."
    , default=False, is_flag=True
)
@click.option('-p', '--package'
    , help="Create a Python project with a package structure rather than a module structure:\n\n"
           "* package structure = ``<module_name>/__init__.py``\n"
           "* module  structure = ``<module_name>.py`` \n"
    , default=False, is_flag=True
)
@click.option('--micc-file'
    , help="The file containing the descriptions of the template parameters used"
           "in the *Cookiecutter* templates. "
    , default='', type=Path
)
@click.option('--python'
    , help="minimal python version for your project."
    , default='3.7'
)
@click.option('-d', '--description'
    , help="Short description of your project."
    , default='<Enter a one-sentence description of this project here.>'
)
@click.option('-l', '--lic'
    , help="License identifier."
    , default='MIT'
)
@click.option('-T', '--template', help=__template_help, default=[])
@click.option('-n', '--allow-nesting'
    , help="If specified allows to nest a project inside another project."
    , default=False, is_flag=True
)
@click.option('--module-name'
    , help="use this name for the module, rather than deriving it from the project name."
    , default=''
)
@click.option('--remote'
    , help="Create remote repo on github. 'public'*|'private'|'none'"
    , default='public'
)
@click.argument('name', type=str, default='')
@click.pass_context
def create(ctx
           , name
           , package
           , module_name
           , micc_file
           , description
           , python
           , lic
           , template
           , allow_nesting
           , publish
           , remote
           ):
    """Create a new project skeleton.

    The project name is taken to be the last directory of the *project_path*.
    If this directory does not yet exist, it is created. If it does exist already, it
    must be empty.

    The package name is the derived from the project name, taking the
    `PEP8 module naming rules <https://www.python.org/dev/peps/pep-0008/#package-and-module-names>`_
    into account:

    * all lowercase.
    * dashes ``'-'`` and spaces ``' '`` replaced with underscores ``'_'``.
    * in case the project name has a leading number, an underscore is prepended ``'_'``.

    If *project_path* is a subdirectory of a micc project, *micc* refuses to continu,
    unless ``--allow-nesting`` is soecified.
    """
    options = ctx.obj

    if name:
        if not options.default_project_path:
            # global option -p and argument name were both specified.
            print( "ERROR: you specified both global option -p and argument 'name':"
                  f"         -p -> {options.project_path}"
                  f"         name -> {name}"
                   "       You must choose one or the other, not both."
                 )
            ctx.exit(-1)
        else:
            # overwrite the -p global option so the project will be created:
            options.project_path = Path(name).resolve()
            options.default_project_path = False

    options.create = True
    options.micc_file = micc_file
    options.package = package
    options.publish = publish
    options.module_name = module_name
    if not remote in ['public','private']:
        remote = None
    options.remote = remote

    if not template:  # default, empty list
        if options.package:
            template = [ 'package-base'
                       , 'package-general'
                       , 'package-simple-docs'
                       , 'package-general-docs'
                       ]
        else:
            template = [ 'package-base'
                       , 'package-simple'
                       , 'package-simple-docs'
                       ]
        options.templates = template
    # else:
    #     # ignore structure
    #     options.structure = 'user-defined'

    options.allow_nesting = allow_nesting

    licenses = ['MIT license'
        , 'BSD license'
        , 'ISC license'
        , 'Apache Software License 2.0'
        , 'GNU General Public License v3'
        , 'Not open source'
                ]
    for l in licenses:
        if l.startswith(lic):
            license_ = l
            break
    else:
        license_ = licenses[0]

    options.template_parameters.update(
        {'project_short_description': description,
         'open_source_license': license_,
         'python_version': python,
        }
    )
    project = Project(options)
    if project.exit_code:
        ctx.exit(project.exit_code)


@main.command()
@click.option('--overwrite', is_flag=True
    , help="Overwrite pre-existing files (without backup)."
    , default=False
)
@click.option('--backup', is_flag=True
    , help="Make backup files (.bak) before overwriting any pre-existing files."
    , default=False
)
@click.pass_context
def convert_to_package(ctx, overwrite, backup):
    """Convert a Python module project to a package.

    A Python *module* project has only a ``<package_name>.py`` file, whereas
    a Python *package* project has ``<package_name>/__init__.py`` and can contain
    submodules, such as Python modules, packages and applications, as well as
    binary extension modules.

    This command also expands the ``package-general-docs`` template in this
    project, which adds a ``AUTHORS.rst``, ``HISTORY.rst`` and ``installation.rst``
    to the documentation structure.
    """
    options = ctx.obj
    options.overwrite = overwrite
    options.backup = backup

    project = Project(options)
    if project.exit_code:
        ctx.exit(project.exit_code)

    with et_micc.logger.logtime(options):
        project.module_to_package_cmd()

        if project.exit_code == et_micc.expand.EXIT_OVERWRITE:
            options.logger.warning(
                f"It is normally ok to overwrite 'index.rst' as you are not supposed\n"
                f"to edit the '.rst' files in '{options.project_path}{os.sep}docs.'\n"
                f"If in doubt: rerun the command with the '--backup' flag,\n"
                f"  otherwise: rerun the command with the '--overwrite' flag,\n"
            )

    if project.exit_code:
        ctx.exit(project.exit_code)


@main.command()
@click.option('--name', is_flag=True
    , help="print the project name."
    , default=False
)
@click.option('--version', is_flag=True
    , help="print the project version."
    , default=False
)
@click.pass_context
def info(ctx,name,version):
    """Show project info.

    * file location
    * name
    * version number
    * structure (with ``-v``)
    * contents (with ``-vv``)

    Use verbosity to produce more detailed info.
    """
    options = ctx.obj

    project = Project(options)
    if project.exit_code:
        ctx.exit(project.exit_code)

    if name:
        print(project.package_name)
        return
    if version:
        print(project.version)
        return
    else:
        with et_micc.logger.logtime(options):
            project.info_cmd()

    if project.exit_code:
        ctx.exit(project.exit_code)


@main.command()
@click.option('-M', '--major'
    , help='Increment the major version number component and set minor and patch components to 0.'
    , default=False, is_flag=True
)
@click.option('-m', '--minor'
    , help='Increment the minor version number component and set minor and patch component to 0.'
    , default=False, is_flag=True
)
@click.option('-p', '--patch'
    , help='Increment the patch version number component.'
    , default=False, is_flag=True
)
@click.option('-r', '--rule'
    , help='Any semver 2.0 version string.'
    , default=''
)
@click.option('-t', '--tag'
    , help='Create a git tag for the new version, and push it to the remote repo.'
    , default=False, is_flag=True
)
@click.option('-s', '--short'
    , help='Print the version on stdout.'
    , default=False, is_flag=True
)
@click.option('-d', '--dry-run'
    , help='bumpversion --dry-run.'
    , default=False, is_flag=True
)
@click.pass_context
def version(ctx, major, minor, patch, rule, tag, short, dry_run):
    """Modify or show the project's version number."""
    options = ctx.obj

    if rule and (major or minor or patch):
        msg = ("Both --rule and --major|--minor|--patc specified.")
        click.secho("[ERROR]\n" + msg, fg='bright_red')
        ctx.exit(1)
    elif major:
        rule = 'major'
    elif minor:
        rule = 'minor'
    elif patch:
        rule = 'patch'

    options.rule = rule
    options.short = short
    options.dry_run = dry_run

    project = Project(options)
    if project.exit_code:
        ctx.exit(project.exit_code)

    with et_micc.logger.logtime(project):
        project.version_cmd()
        if project.exit_code == 0 and tag:
            project.tag_cmd()

    if project.exit_code:
        ctx.exit(project.exit_code)


@main.command()
@click.pass_context
def tag(ctx):
    """Create a git tag for the current version and push it to the remote repo."""
    options = ctx.obj

    project = Project(options)
    if project.exit_code:
        ctx.exit(project.exit_code)

    if project.exit_code:
        ctx.exit(project.exit_code)

    project.tag_cmd()

    if project.exit_code:
        ctx.exit(project.exit_code)


@main.command()
@click.option('--app'
    , default=False, is_flag=True
    , help="Add a CLI ."
)
@click.option('--group'
    , default=False, is_flag=True
    , help="Add a CLI with a group of sub-commands rather than a single command CLI."
)
@click.option('--py'
    , default=False, is_flag=True
    , help="Add a Python module."
)
@click.option('--package'
    , help="Add a Python module with a package structure rather than a module structure:\n\n"
           "* module  structure = ``<module_name>.py`` \n"
           "* package structure = ``<module_name>/__init__.py``\n\n"
           "Default = module structure."
    , default=False, is_flag=True
)
@click.option('--f90'
    , default=False, is_flag=True
    , help="Add a f90 binary extionsion module (Fortran)."
)
@click.option('--cpp'
    , default=False, is_flag=True
    , help="Add a cpp binary extionsion module (C++)."
)
@click.option('-T', '--templates', default='', help=__template_help)
@click.option('--overwrite', is_flag=True
    , help="Overwrite pre-existing files (without backup)."
    , default=False
)
@click.option('--backup', is_flag=True
    , help="Make backup files (.bak) before overwriting any pre-existing files."
    , default=False
)
@click.argument('name', type=str)
@click.pass_context
def add(ctx
        , name
        , app, group
        , py, package
        , f90
        , cpp
        , templates
        , overwrite
        , backup
        ):
    """Add a module or CLI to the projcect.

    :param str name: name of the CLI or module added.

    If ``app==True``: (add CLI application)

    * :py:obj:`app_name` is also the name of the executable when the package is installed.
    * The source code of the app resides in :file:`<project_name>/<package_name>/cli_<name>.py`.


    If ``py==True``: (add Python module)

    * Python source  in :file:`<name>.py*`or :file:`<name>/__init__.py`, depending on the :py:obj:`package` flag.

    If ``f90==True``: (add f90 module)

    * Fortran source in :file:`f90_<name>/<name>.f90` for f90 binary extension modules.

    If ``cpp==True``: (add cpp module)

    * C++ source     in :file:`cpp_<name>/<name>.cpp` for cpp binary extension modules.
    """
    options = ctx.obj
    options.add_name = name
    options.app = app
    options.group = group
    options.py = py
    options.package = package
    options.f90 = f90
    options.cpp = cpp
    options.templates = templates
    options.overwrite = overwrite
    options.backup = backup

    project = Project(options)
    if project.exit_code:
        ctx.exit(project.exit_code)

    with et_micc.logger.logtime(options):
        project.add_cmd()

    if project.exit_code:
        ctx.exit(project.exit_code)


@main.command()
@click.option('--silent', is_flag=True
    , help="Do not ask for confirmation on deleting a component."
    , default=False
)
@click.option('--entire-package', is_flag=True
    , help="Replace all occurences of <cur_name> in the entire package and in the ``tests`` directory."
    , default=False
)
@click.option('--entire-project', is_flag=True
    , help="Replace all occurences of <cur_name> in the entire project."
    , default=False
)
@click.argument('cur_name', type=str)
@click.argument('new_name', type=str, default='')
@click.pass_context
def mv(ctx, cur_name, new_name, silent, entire_package, entire_project):
    """Rename or remove a component, i.e an app (CLI) or a submodule.

    :param cur_name: name of component to be removed or renamed.
    :param new_name: new name of the component. If empty, the component will be removed.
    """
    options = ctx.obj

    options.cur_name = cur_name
    options.new_name = new_name
    options.silent = silent
    if new_name:
        options.entire_package, options.entire_project =  entire_package, entire_project
    # else these flags are ignored.

    project = Project(options)
    if project.exit_code:
        ctx.exit(project.exit_code)

    with et_micc.logger.logtime(options):
        project.mv_component()


@main.command()
@click.option('--force', '-f', is_flag=True
    , help="Overwrite existing setup."
    , default=False
)
@click.pass_context
def setup(ctx
        , force
        ):
    """Setup your micc preferences.

    This command must be run once before you can use micc to manage your projects.
    """
    options = ctx.obj

    options.force = force
    micc_file_template = Path(__file__).parent / 'micc.json'
    setupdir = Path().home() / '.et_micc'
    setupdir.mkdir(exist_ok=True)
    path_to_miccfile = setupdir / 'micc.json'
    if not path_to_miccfile.exists() or force:
        shutil.copyfile(str(micc_file_template),str(path_to_miccfile))
        preferences = et_micc.expand.set_preferences(path_to_miccfile)
        print("\nConfiguring git:")
        cmds = [['git', 'config', '--global', 'user.name' , preferences['full_name']['default'] ]
               ,['git', 'config', '--global', 'user.email', preferences['email'    ]['default'] ]
               ,['git', 'config', '--global', 'credential.helper', 'cache']
               ]
        et_micc.utils.execute(cmds, print, stop_on_error=False)

    else:
        print("Micc has already been setup. Use '--force' or '-f' to overwrite the existing setup.")
    print("\nIf you want to change your preferences, edit the default entries in file \n"
         f"    {path_to_miccfile}\n"
          "Note that these changes will only affect NEW projects. Existing projects will be unaffected.\n"
          "Micc is now configured and ready to be used.\n\n"
          "Done"
    )

if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover

# eof



# -*- coding: utf-8 -*-
"""
Module et_micc.db
=================

This submodule implements the database that is needed for safe refactoring of
submodules and CLIs.
"""

class Database:
    def __init__(self,project):
        self.project = project

    def add_app(self):
        """Database action when an app (CLI) is added."""

    def add_python_module(self):
        """Database action when a python module is added."""

    def add_f90_module(self):
        """Database action when a f90 module is added."""

    def add_cpp_module(self):
        """Database action wehn a C++ module is added."""

    def remove_app(self):
        """Database action when an app (CLI) is removed."""

    def remove_python_module(self):
        """Database action when a python module is removed."""

    def remove_f90_module(self):
        """Database action when a f90 module is removed."""

    def remove_cpp_module(self):
        """Database action wehn a C++ module is removed."""

    def rename_app(self):
        """Database action when an app (CLI) is renamed."""

    def rename_python_module(self):
        """Database action when a python module is renamed."""

    def rename_f90_module(self):
        """Database action when a f90 module is renamed."""

    def rename_cpp_module(self):
        """Database action wehn a C++ module is renamed."""




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['et_micc']

package_data = \
{'': ['*'],
 'et_micc': ['templates/app-simple/{{cookiecutter.project_name}}/tests/*',
             'templates/app-simple/{{cookiecutter.project_name}}/{{cookiecutter.package_name}}/*',
             'templates/app-sub-commands/{{cookiecutter.project_name}}/tests/*',
             'templates/app-sub-commands/{{cookiecutter.project_name}}/{{cookiecutter.package_name}}/*',
             'templates/module-cpp/{{cookiecutter.project_name}}/tests/*',
             'templates/module-cpp/{{cookiecutter.project_name}}/{{cookiecutter.package_name}}/cpp_{{cookiecutter.module_name}}/*',
             'templates/module-f90/{{cookiecutter.project_name}}/tests/*',
             'templates/module-f90/{{cookiecutter.project_name}}/{{cookiecutter.package_name}}/f90_{{cookiecutter.module_name}}/*',
             'templates/module-py/{{cookiecutter.project_name}}/tests/*',
             'templates/module-py/{{cookiecutter.project_name}}/{{cookiecutter.package_name}}/*',
             'templates/package-base/hooks/*',
             'templates/package-base/{{cookiecutter.project_name}}/*',
             'templates/package-base/{{cookiecutter.project_name}}/tests/*',
             'templates/package-general-docs/hooks/*',
             'templates/package-general-docs/{{cookiecutter.project_name}}/*',
             'templates/package-general-docs/{{cookiecutter.project_name}}/docs/*',
             'templates/package-general/hooks/*',
             'templates/package-general/{{cookiecutter.project_name}}/{{cookiecutter.package_name}}/*',
             'templates/package-simple-docs/hooks/*',
             'templates/package-simple-docs/{{cookiecutter.project_name}}/*',
             'templates/package-simple-docs/{{cookiecutter.project_name}}/docs/*',
             'templates/package-simple/hooks/*',
             'templates/package-simple/{{cookiecutter.project_name}}/*']}

install_requires = \
['click>=7.0,<8.0',
 'cookiecutter>=1.6.0,<2.0.0',
 'pypi-simple>=0.8.0,<0.9.0',
 'semantic_version>=2.8.3,<3.0.0',
 'sphinx-click>=2.3.0,<3.0.0',
 'sphinx-rtd-theme>=0.4.3,<0.5.0',
 'tomlkit>=0.5.8,<0.6.0',
 'walkdir>=0.4.1,<0.5.0']

entry_points = \
{'console_scripts': ['micc = et_micc:cli_micc.main']}

setup_kwargs = {
    'name': 'et-micc',
    'version': '1.1.8',
    'description': 'A practical Python project skeleton generator.',
    'long_description': "****\nMicc\n****\n\n.. image:: https://img.shields.io/pypi/v/micc.svg\n        :target: https://pypi.python.org/pypi/micc\n\n.. image:: https://img.shields.io/travis/etijskens/micc.svg\n        :target: https://travis-ci.org/etijskens/micc\n\n.. image:: https://readthedocs.org/projects/micc/badge/?version=latest\n        :target: https://micc.readthedocs.io/en/latest/?badge=latest\n        :alt: Documentation Status\n\n\n`Micc <https://github.com/etijskens/et-micc>`_ is a Python project manager: it helps \nyou organize your Python project from simple single file modules to fully fledged \nPython packages containing modules, sub-modules, apps and binary extension modules \nwritten in Fortran or C++. Micc_ organizes your project in a way that is considered good\npractice by a large part of the Python community. \n\n* Micc_ helps you create new projects. You can start small with a simple one-file \n  package and add material as you go, such as:\n  \n  * Python **sub-modules** and **sub-packages**,\n  * **applications**, also known as command line interfaces (CLIs). \n  * **binary extension modules** written in C++ and Fortran. Boiler plate code is \n    automatically added as to build these binary extension with having to go through\n    al the details. This is, in fact, the foremost reason that got me started on this\n    project: For *High Performance Python* it is essential to rewrite slow and \n    time consuming parts of a Python script or module in a language that is made \n    for High Performance Computing. As figuring out how that can be done, requires \n    quite some effort, Micc_ was made to automate this part while maintaining the \n    flexibility. \n  * Micc_ adds typically files containing example code to show you how to add your\n    own functionality.\n    \n* You can automatically **extract documentation** from the doc-strings of your files,\n  and build html documentation that you can consult in your browser, or a .pdf \n  documentation file.\n* With a little extra effort the generated html **documentation is automatically published** \n  to `readthedocs <https://readthedocs.org>`_.\n* Micc_ helps you with **version management and control**.\n* Micc_ helps you with **testing** your code.\n* Micc_ helps you with **publishing** your code to e.g. `PyPI <https://pypi.org>`_, so\n  that you colleagues can use your code by simply running::\n\n    > pip install your_nifty_package\n\nCredits\n-------\nMicc_ does not do all of this by itself. For many things it relies on other strong \nopen source tools and it is therefor open source as well (MIT Licence). Here is a list \nof tools micc_ is using or cooperating with happily:\n\n*   `Pyenv <https://github.com/pyenv/pyenv>`_: management of different Python versions.\n*   `Pipx <https://github.com/pipxproject/pipx/>`_ for installation of CLIs in a system-wide\n    way.\n*   `Poetry <https://github.com/sdispater/poetry>`_ for dependency management, virtual\n    environment management, packaging and publishing.\n*   `Git <https://www.git-scm.com/>`_ for version control.\n*   `CMake <https://cmake.org>`_ is usde for building binary extension modules written\n    in C++.\n\nThe above tools are not dependencies of Micc_ and must be installed separately. Then\nthere are a number of python packages on which micc_ depends and which are automatically\ninstalled when poetry_ creates a virtual environment for a project.\n\n*   `Cookiecutter <https://github.com/audreyr/cookiecutter>`_ for creating boilerplate\n    code from templates for all the parts that can be added to your project.\n*   `Python-semanticversion <https://github.com/rbarrois/python-semanticversion/blob/master/docs/index.rst>`_\n    for managing version strings and dependency version constraints according to the\n    `Semver 2.0 <http://semver.org/>`_ specification.\n*   `Pytest <https://www.git-scm.com/>`_ for testing your code.\n*   `Click <https://click.palletsprojects.com/en/7.x/>`_ for a pythonic and intuitive definition\n    of command-line interfaces (CLIs).\n*   `Sphinx <http://www.sphinx-doc.org/>`_ to extract documentation from your project's\n    doc-strings.\n*   `Sphinx-click <https://sphinx-click.readthedocs.io/en/latest/>`_ for extracting documentation\n    from the click_ command descriptions.\n*   `F2py <https://docs.scipy.org/doc/numpy/f2py/>`_ for transforming modern Fortran code into performant\n    binary extension modules interfacing nicely with `Numpy <https://numpy.org/>`_.\n*   `Pybind11 <https://pybind11.readthedocs.io/en/stable/>`_ as the\n    glue between C++ source code and performant binary extension modules, also interfacing nicely with Numpy_.\n\nRoadmap\n=======\nThese features are still on our wish list:\n\n* Contininous integtration (CI)\n* Code style, e.g. `flake8 <http://flake8.pycqa.org/en/latest/>`_ or `black <https://github.com/psf/black>`_\n* Profiling\n* Gui for debugging C++/Fortran binary extensions\n* Micc projects on Windows (So far, only support on Linux and MacOS).\n\n",
    'author': 'Engelbert Tijskens',
    'author_email': 'engelbert.tijskens@uantwerpen.be',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/etijskens/et-micc',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
